/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 23.06.2003
 *
 *
 *
 */
package at.spardat.xma.session;

import java.io.*;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import javax.security.auth.Subject;
import javax.servlet.ServletException;
import javax.servlet.http.*;

import at.spardat.enterprise.exc.SysException;
import at.spardat.properties.PropertyUnknownException;
import at.spardat.properties.XProperties;
import at.spardat.xma.RuntimeDefaults;
import at.spardat.xma.baserpc.BaseRPCServer;
import at.spardat.xma.baserpc.ClientToServerData;
import at.spardat.xma.baserpc.IRPCExecutorServer;
import at.spardat.xma.baserpc.ServerToClientData;
import at.spardat.xma.boot.server.ServerAppLoader;
import at.spardat.xma.plugins.PluginManagerServer;
import at.spardat.xma.security.LoginModuleServer;
import at.spardat.xma.security.LoginModuleServerWithContextChange;
import at.spardat.xma.util.Version;

/**
 * This class handles the server side creation and destroying of sessions.
 *
 * This servlet only services POST-Requests and takes one parameter
 * named "operation". If this parameter is "checkVersion", it sends back the
 * hashcode of the application and the version number of the xma runtime as
 * http servlet response. In all other cases, the reqeust is executed as base rpc.
 * Valid RPCs are:
 * <ul>
 * <li>getChallange: sends back a challenge string for use by the encryption
 * <li>login: validates the user information and creates the {@link XMASessionServer} on success
 * <li>keepAlive: does nothing
 * <li>logout: destroys the HttpSession
 * <li>contextChanged: validates and changes the user information in the {@link XMASessionServer} on success
 * </ul>
 *
 * @author s2877
 */
public class LoginServlet extends HttpServlet implements IRPCExecutorServer {

    /**
     * Handles the post-requests to this servlet. If the request-parameter "operation" is
     * "checkVersion" it sends back the hashcode of the application and the version number
     * of the xma runtime. In all other cases, the reqeust is interpreted as
     * base rpc and executed accordingly.
     * @param request
     * @param response
     * @throws IOException passed from the stream operations
     */
    protected void doPost (HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        String operation = request.getParameter("operation");
        if("checkVersion".equals(operation)) {
            getVersion(request,response);
//        } else if("logout".equals(operation)) {
//            HttpSession session = request.getSession(false);
//            if(session!=null) {
//                session.invalidate();
//            }
        } else {
            new BaseRPCServer().handlePost(request,response,getServletConfig(),this);
        }
    }

    /**
     * Excecutes the given operation of the base rpc. <br/>
     * The valid operation names are:
     * <ul>
     * <li>getPreLoginInfo: sends back pre login information
     * <li>login: validates and completes the user information and creates the {@link XMASessionServer} on success
     * <li>keepAlive: does nothing
     * <li>logout: destroys the HttpSession
     * <li>contextChanged: validates and changes the user information in the {@link XMASessionServer} on success
     * </ul>
     *
     * @param request data send from the client
     * @param response data to be send to the client
     * @param httpRequest original http servlet request
     */
    public void execute(String operation, ClientToServerData request, ServerToClientData response, HttpServletRequest httpRequest) {
        if("getPreLoginInfo".equals(operation)) {
            getPreLoginInfo(request,response,httpRequest.getSession());
        } else if("login".equals(operation)) {
            login(request,response,httpRequest);
        } else if("keepAlive".equals(operation)) {
            httpRequest.getSession();
        } else if("logout".equals(operation)) {
            HttpSession session = httpRequest.getSession();
            if(session!=null) {
                session.invalidate();
            }
        } else if("contextChanged".equals(operation)) {
            contextChanged(request,response,httpRequest.getSession());
        } else {
            throw new SysException("no such rpc: "+operation);
        }
    }

    /**
     * Sends back the pre-login-information gained by calling getPreLoginInfo() on the
     * configured plugin {@link LoginModuleServer}.
     * @param request
     * @param response
     */
    private void getPreLoginInfo(ClientToServerData request, ServerToClientData response, HttpSession session) {
        // handle pre-login information from plugin
        LoginModuleServer module = (LoginModuleServer)PluginManagerServer.getInstance().getPlugin(LoginModuleServer.class);
        //retrieve parameters set in XMAClientSession.serverGetPreLoginInfo()
        String version = (String) request.getParameter(0);
        String ipAdress = (String) request.getParameter(1);
        //set the client ip / version in http session
        if(version != null){
            session.setAttribute(XMASessionServer.CLIENT_BOOTRUNTIME_VERSION,version);
        }
        if(ipAdress != null){
            session.setAttribute(XMASessionServer.CLIENT_IP_ADDRESS,ipAdress);
        }

        Object info = module.getPreLoginInfo(session);
        if(info!=null) {
            response.setParameter(0,new Boolean(true));
            response.setParameter(1,info);
        } else {
            response.setParameter(0,new Boolean(false));
        }
        // handle properties for client side
        Properties props = new Properties();
        XProperties     node = XProperties.getNodeOfPackage(RuntimeDefaults.clientPraefix);
        Iterator        iter = node.getKeys();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            try {
                props.setProperty(key,node.get(key));
             } catch (PropertyUnknownException e) {
                throw new SysException(e);
            }
        }
        for(Iterator it=RuntimeDefaults.keySet().iterator();it.hasNext();) {
            String key = (String) it.next();
            if(key.startsWith(RuntimeDefaults.clientPraefix)) {
                String shortName = key.substring(key.lastIndexOf('.')+1);
                if(!props.containsKey(shortName)) {
                    props.setProperty(shortName,RuntimeDefaults.getProperty(key));
                }
            }
        }
        StringBuffer buffer = new StringBuffer();
        for(Iterator it=props.keySet().iterator();it.hasNext();) {
            String key = (String) it.next();
            buffer.append(key).append('\n');
            buffer.append(props.getProperty(key)).append('\n');
        }
        response.setParameter(2,buffer.toString());
    }

    /**
     * Checks the user information send by the client and authenticates the HttpSession.
     * The user information read by deserialising the principals and publicCredentials sets
     * of the javax.security.auth.Subject from the request body. The authentication is done
     * by calling login() on the configured plugin {@link LoginModuleServer}.
     * If the authentication succeeds a new {@link XMASessionServer} which contains the user
     * information is created. The pricipals and publicCredentials are serialized back
     * to the client in any case, so the server side login plugin can transfer additional
     * information to the client side.<br>
     * Sends back to the client:<br>
     * if the login succees: true and the principals and publicCredentials<br>
     * it the login fails: false and the principals and publicCredentials<br>
     * nothing in case of technical problems<br>
     *
     * @param request
     * @param response
     */
    private void login(ClientToServerData request, ServerToClientData response, HttpServletRequest httpRequest) {
        Set principals = new HashSet();
        principals.addAll(request.getUnserializableCollection(0));
        Set credentials = new HashSet();
        credentials.addAll(request.getUnserializableCollection(1));
        Subject subject = new Subject(false,principals,credentials,new HashSet());
        LoginModuleServer module = (LoginModuleServer)PluginManagerServer.getInstance().getPlugin(LoginModuleServer.class);

        HttpSession session = httpRequest.getSession();
        boolean ok = module.login(subject,session);
        if(ok) {
            subject.setReadOnly();
            // treat context path
            String      contextPath = httpRequest.getContextPath();
            if (contextPath.length() > 0 && contextPath.charAt(0) == '/')
                contextPath = contextPath.substring(1);
            new XMASessionServer(session, subject, contextPath);
            response.setParameter(0,new Boolean(true));
        } else {
            response.setParameter(0,new Boolean(false));
        }
        response.setUnserializableCollection(1,subject.getPrincipals());
        response.setUnserializableCollection(2,subject.getPublicCredentials());
    }

    /**
     * sends back the hashcode of the application
     * @param request
     * @param response
     * @throws IOException
     */
    private void getVersion(HttpServletRequest request, HttpServletResponse response) throws IOException {
        OutputStream out = response.getOutputStream();
        out.write( ServerAppLoader.getDigest() );
        out.write(Version.getVersion().getBytes("US-ASCII"));
    }

    private void contextChanged(ClientToServerData request, ServerToClientData response, HttpSession httpsession) {
        Set principals = new HashSet();
        principals.addAll(request.getUnserializableCollection(0));
        Set credentials = new HashSet();
        credentials.addAll(request.getUnserializableCollection(1));
        Subject subject = new Subject(false,principals,credentials,new HashSet());
        LoginModuleServerWithContextChange module = (LoginModuleServerWithContextChange)PluginManagerServer.getInstance().getPlugin(LoginModuleServer.class);
        XMASessionServer session = (XMASessionServer) httpsession.getAttribute(XMASessionServer.SESSION_ATTRIBUTE_KEY);
        if(session!=null) {
            if(module.contextChanged(subject,httpsession)) {
                subject.setReadOnly();
                session.setSubject(subject);
                response.setParameter(0,new Boolean(true));
            } else {
                httpsession.invalidate();
                response.setParameter(0,new Boolean(false));
            }
        } else {
            response.setParameter(0,new Boolean(false));
        }
        response.setUnserializableCollection(1,subject.getPrincipals());
        response.setUnserializableCollection(2,subject.getPublicCredentials());
    }

}
